
package org.glassfish.websockets.platform;

import org.glassfish.grizzly.websockets .*;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import org.glassfish.websockets.platform.processors.WSProcessor;

/**
 *
 * @author dannycoward
 */

public class BeanServer {
    static String hostname = "localhost"; // default server settings
    
    final static Logger logger = Logger.getLogger("wsplatform");
    List<WebSocketApplication> applications = new ArrayList<WebSocketApplication>();
    private static ContainerContextImpl containerContext;

    public static ContainerContextImpl getContainerContext() {
        return containerContext;
    }
    
    public void closeWebSocketServer() {
        for (WebSocketApplication wsa : this.applications) {
            ((WebSocketEndpoint) wsa).remove();
            WebSocketEngine.getEngine().unregister(wsa);
            logger.info("Closing down : " + wsa);
        }
    }
    
    public void initWebSocketServer(String wsPath, int port, Set<Class<?>> fqWSBeanNames) throws Exception {
        this.containerContext = new ContainerContextImpl(this, wsPath, port);
        for (Class<?> webSocketApplicationBeanClazz : fqWSBeanNames) {
            String fqWSBeanName = webSocketApplicationBeanClazz.getName();
            Object webSocketApplicationBean = webSocketApplicationBeanClazz.newInstance();
        
            logger.info("Loaded application bean " + fqWSBeanName + " ..." + webSocketApplicationBean);
        
            // introspect the bean and find all the paths....
            Map methodPathMap = this.getMethodToPathMap(webSocketApplicationBeanClazz);
            if (methodPathMap.isEmpty()) {
                logger.warning(webSocketApplicationBeanClazz + " has no path mappings");
            }
            WebSocketApplication applicationBeanWrapper = null;
        
            Set<String> allPathsForBean = new HashSet(methodPathMap.values());
            
            // create one adapter per path. So each class may have multiple adapters.
            for (String nextPath : allPathsForBean) {
                String wrapperClassname = WSProcessor.getQualifiedClassnameFor(fqWSBeanName, nextPath);
                Set<Method> allMethodsForPath = getMethodsForPath(webSocketApplicationBeanClazz, nextPath);
                
                // Use the endpoint's classloader to load wrapper classes
                Class wrapperClass = webSocketApplicationBeanClazz.getClassLoader().loadClass(wrapperClassname);
                
                applicationBeanWrapper = (WebSocketApplication) wrapperClass.newInstance();
            
                //applicationBeanWrapper = new Manual_BeanDispatcher();
            
                //applicationBeanWrapper = new wsgenerated.BeanWrapper(); // NetBeans does barf on this
                String wrapperBeanPath = wsPath+ nextPath;
                ((Initable) applicationBeanWrapper).doInit(wrapperBeanPath, webSocketApplicationBean, allMethodsForPath);
                logger.info("Created wrapper bean: " + applicationBeanWrapper);
                this.deploy(applicationBeanWrapper);
            }
        }
    }
    
    void deploy(WebSocketApplication wsa) {
        WebSocketEngine.getEngine().register(wsa);
        this.applications.add(wsa);
        logger.info("Registered a " + wsa.getClass() + " at " + ((WebSocketEndpoint) wsa).getPath());
    }
    
    private Set getMethodsForPath(Class beanClazz, String path) throws Exception {
        Set<Method> s = new HashSet<Method>();
        Map<Method, String> methodPath = this.getMethodToPathMap(beanClazz);
        for (Method m : methodPath.keySet()) {
            String p = methodPath.get(m);
            if (p.equals(path)) {
                s.add(m);
            }
        }
        return s;
    }
    
    
    private Map<Method, String> getMethodToPathMap(Class beanClazz) throws Exception {
        Map<Method, String> pathMappings = new HashMap();
        Method[] methods = beanClazz.getDeclaredMethods();
        for (Method method : methods) {
            org.glassfish.websockets.api.annotations.WebSocket wsClass = (org.glassfish.websockets.api.annotations.WebSocket) beanClazz.getAnnotation(org.glassfish.websockets.api.annotations.WebSocket.class);
            pathMappings.put(method, wsClass.path());
        }
        return pathMappings;  
    }
    
 
    

}
