
package org.glassfish.websockets.platform;

import org.glassfish.grizzly.websockets.*;
import java.io.*;
import org.glassfish.websockets.api.*;
import java.util.*;
import java.util.logging.*;
import java.net.*;
/**
 *
 * @author dannycoward
 */
public class WebSocketWrapper implements Peer {     
    private WebSocket grizzlySocket;
    private WebSocketConversationImpl webSocketSession;
    private EndpointContextImpl webSocketContext;
    private String clientAddress;
    Date activationTime = new Date();
    private static Set<WebSocketWrapper> wsws = new HashSet<WebSocketWrapper>();
    
    final static Logger logger = Logger.getLogger("wsplatform");

    public static WebSocketWrapper getWebSocketWrapper(WebSocket socket, WebSocketEndpoint application) {
        String remoteImpl = application.getRemoteImplClassname();
        //logger.info("getting wrapper for socket and type " + remoteImpl + " for " + application);

        WebSocketWrapper wsw = null;
        if ( WebSocketWrapper.findWebSocketWrapper(socket) != null) {
            // there is one with the same socket 
            // but is it of the right type ?
            wsw = WebSocketWrapper.findWebSocketWrapper(socket);
            if (wsw.getClass().getCanonicalName().equals(remoteImpl)) {
                //logger.info("Reusing old wsw");
            } else {
                throw new RuntimeException("Should not get here. You are asking for a remote impl :"+remoteImpl+": of a different type from one I already have :"+wsw.getClass().getCanonicalName()+": !");
            }

        } else { // create a new one
            if (remoteImpl.equals(WebSocketWrapper.class.getCanonicalName())) {
                wsw = new WebSocketWrapper();
            } else {
                try {
                    Class c = Class.forName(remoteImpl);
                    wsw = (WebSocketWrapper) c.newInstance();
                } catch (Exception e) {
                    throw new RuntimeException("Cannot instantiate a " + remoteImpl);
                }
            }
            wsw.grizzlySocket = socket;
            wsw.webSocketSession = new WebSocketConversationImpl(wsw);
            wsw.webSocketContext = application.getWebSocketContext();
            wsw.webSocketContext.addWebSocketSession(wsw.webSocketSession);
            
            wsws.add(wsw);   
        }
        //logger.info("got the wsw - done..." + wsw.getClass());
        return wsw;

    }
    
    private static Set<WebSocketWrapper> getWebSocketWrappers() {
        weedExpiredWebSocketWrappers();
        return wsws;
    }
    
    private static void weedExpiredWebSocketWrappers() {
        Set<WebSocketWrapper> expired = new HashSet<WebSocketWrapper>();
        for (WebSocketWrapper wsw : wsws) {
            if (!wsw.isConnected()) {
                expired.add(wsw);
            }
        }
        for (WebSocketWrapper toRemove : expired) {
            wsws.remove(toRemove);
        }
    }
    
    static WebSocketWrapper findWebSocketWrapper(WebSocket socket) {
        for (WebSocketWrapper nextWsw : getWebSocketWrappers()) {
            if (nextWsw.grizzlySocket == socket) {
                return nextWsw;
            }
        }
        return null;
    }
    
    // may be called by subclasses...
    protected WebSocketWrapper() {
        this.activationTime = new Date();
    }
    
    public String getAddress() {
        return this.clientAddress;
    }
    
    void setAddress(String clientAddress) {
        this.clientAddress = clientAddress;
    }

    
    public EndpointContext getContext() {
        return this.webSocketContext;
    }
    
    public Conversation getConversation() {
        return this.webSocketSession;
    }
    
    public boolean isConnected() {
        return this.grizzlySocket.isConnected();
    }
    
    
    public void sendPrimitiveMessage(Object data) throws IOException, ConversionException {
        String dataClassname = data.getClass().getName();
        if (dataClassname.equals("int") ||
             dataClassname.equals("byte") ||
                dataClassname.equals("short") ||
                dataClassname.equals("long") ||
                dataClassname.equals("float") ||
                dataClassname.equals("double") ||
                dataClassname.equals("boolean") ||
                dataClassname.equals("char")) {
            this.sendMessage(data.toString());
        } else {
            throw new ConversionException("object " + data + " is not a primitive type.");
        }
    }
    
    WebSocket getSocket() {
        return this.grizzlySocket;
    }
    
    public void sendMessage(String data) throws IOException {
        this.grizzlySocket.send(data);
    }
    
    public void sendMessage(byte[] data) throws IOException {
        this.grizzlySocket.send(data);
    }
    
    
    public void doNothing() throws ConversionException {
        if (false) throw new ConversionException("");
    }
    
    public String toString() {
        return "WSW: " + getClass().getSimpleName();
    }

    public void doClose(int code, String reason) throws IOException {
        this.grizzlySocket.close();
    }

}
