/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.glassfish.websockets.platform.processors;

import org.glassfish.websockets.api.annotations.WebSocketMessage;
import org.glassfish.websockets.api.annotations.WebSocketClose;
import org.glassfish.websockets.api.annotations.WebSocketOpen;
import org.glassfish.websockets.api.annotations.WebSocketContext;
import org.glassfish.websockets.platform.WebSocketWrapper;
import java.io.IOException;
import java.io.PrintWriter;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import java.util.*;
/**
 *
 * @author dannycoward
 */
public class ClassWriter {
    
    private static Element findContextInstanceVariableElement(TypeElement clazz) {
        for (Element e : clazz.getEnclosedElements()) {
            WebSocketContext wsc = e.getAnnotation(WebSocketContext.class);
            if (wsc != null ) return e;
        }
        return null;
    }
    
    public static void writeClass(String packageNameToWrite, 
                                    String clazzNameToWrite, 
                                    TypeElement clazz, 
                                    Set<Element> methods, 
                                    RemoteProcessor rp, 
                                    PrintWriter pw) throws IOException {
        pw.println("package " + packageNameToWrite + ";");
        pw.println();
        ClassWriter.writeImports(pw, clazz);
        pw.println("public class "
                + clazzNameToWrite + " extends WebSocketEndpoint implements Initable {");

        pw.println("\tprivate " + clazz.getSimpleName() + " bean;");
        //pw.println("\tprivate String subPath;");

        pw.println();
        pw.println("\tpublic " + clazzNameToWrite + "() {");
        pw.println("\t}");
        pw.println();

        Element contextIVElement = findContextInstanceVariableElement(clazz);
        
        pw.println("\tpublic void doInit(String path, Object bean, Set methods) {");
        pw.println("\t\tsuper.path = path;");
        pw.println("\t\tthis.bean = (" + clazz.getSimpleName() + ") bean;");
        if (contextIVElement != null) {
            pw.println("\t\tthis.bean." + contextIVElement + " = super.webSocketContext;");
        }
        //deal with subprotocols supported by this class
        org.glassfish.websockets.api.annotations.WebSocket wsa = clazz.getAnnotation(org.glassfish.websockets.api.annotations.WebSocket.class);
        String[] subprotocols = wsa.subprotocols();
        if (subprotocols != null) {
            for (int i = 0; i < subprotocols.length; i++) {
                pw.println("\t\tsuper.addSupportedSubprotocol(\""+subprotocols[i]+"\");");
            }
        }
        
        // set the custom remote implementation class, if necessary
        if (!wsa.remote().equals("")) {
            String riclz = rp.getImplementationFor(wsa.remote());
            pw.println("\t\tsuper.remoteImplClassname = \"" + riclz + "\";");
        }
        
        
        
        pw.println("\t}");
        pw.println();
        
        
        
        
        
        pw.println("\tvoid handleException(Exception e) {");
        pw.println("\t\tSystem.out.println(\"Error replying to client \" + e.getMessage());");
        pw.println("\t\te.printStackTrace();");
        MethodWriter.writeExceptionHandlerMethodBody(clazz, methods, pw);

        pw.println("\t}");
        pw.println();
        
        // may be multiple methods asking for onMessage calls, all with the same path.
        List<Element> onMessageMethods = getOnMessageMethods(methods);
        MethodWriter.writeOnMessageMethod(clazz, onMessageMethods, rp, pw);
        
        // write open and close methods
        for (Element nextMethod : methods) {
            if (nextMethod.getAnnotation(WebSocketOpen.class) != null) {
                MethodWriter.writeOnOpenMethod(clazz, nextMethod, rp, pw);
            } else if (nextMethod.getAnnotation(WebSocketClose.class) != null) {
                MethodWriter.writeOnCloseMethod(clazz, nextMethod, rp, pw);
            } 
            pw.println();
        }
        pw.println("}");

    }

    private static List<Element> getOnMessageMethods(Set<Element> methods) {
        List<Element> onMessageMethods = new ArrayList<Element>();
        for (Element nextMethod : methods) {
            if (nextMethod.getAnnotation(WebSocketMessage.class) != null) {
                onMessageMethods.add(nextMethod);
            }
        }
        return onMessageMethods;
    }
    
    public static void writeImports(PrintWriter pw, TypeElement eachClazz) {
        pw.println("import org.glassfish.grizzly.websockets.*;");
        pw.println("import org.glassfish.websockets.platform.*;");
        pw.println("import org.glassfish.websockets.api.*;");
        pw.println("import java.util.*;");
        pw.println("import " + eachClazz.getQualifiedName() + ";");
        pw.println();
        
    }
}
