package org.glassfish.websockets.platform.processors;

import org.glassfish.websockets.api.annotations.*;
import java.io.IOException;
import java.io.PrintWriter;
import javax.lang.model.type.*;
import java.util.*;
import java.lang.reflect.*;

import javax.lang.model.element.*;

/**
 *
 * @author dannycoward
 */
public class MethodWriter {
    
    static String getClassPath(Element clazz) {
        WebSocket ws = clazz.getAnnotation(org.glassfish.websockets.api.annotations.WebSocket.class);
        if (ws != null) {
            return ws.path();
        } else {
            return "";
        }
    }
    
    static void writeExceptionHandlerMethodBody(TypeElement clazz, Set<Element> methods, PrintWriter pw) {
        for (Element method : methods) {
            WebSocketError wse = method.getAnnotation(org.glassfish.websockets.api.annotations.WebSocketError.class);
            if (wse != null) {
                String endPointPath = getClassPath(clazz);
                pw.println("\t\tif (true) {");
                pw.println("\t\t\tbean." + method.getSimpleName() + "(e);");
                pw.println("\t\t}");   
            }   
        }
    }
    
    static boolean checkRemoteMatching(TypeElement clazz, String remoteImpl, RemoteProcessor rp) {
        boolean match = false;
        String classDeclaredRemoteImpl = "";
        WebSocket wsa = clazz.getAnnotation(WebSocket.class);
        String classLevelRemoteIfc = wsa.remote();
        if (!classLevelRemoteIfc.equals("")) {
            classDeclaredRemoteImpl = rp.getImplementationFor(wsa.remote());
            match =  classDeclaredRemoteImpl.equals(remoteImpl);
        }
        if (!match) {
            throw new RuntimeException("Syntax error ?  A method on class :"+clazz.getQualifiedName()+": is looking for an :" + remoteImpl + ": But its class declares :" + classLevelRemoteIfc + ": as a remote ifc with impl is :" + classDeclaredRemoteImpl + ":. ???" );
        }
        return true;
    }
    
    static void writeOnOpenMethod(TypeElement clazz, 
                                    Element method, 
                                    RemoteProcessor rp, 
                                    PrintWriter pw) {
        String openMethodName = method.getSimpleName().toString();
        ExecutableElement eMethod = (ExecutableElement) method;
        
        pw.println("\tpublic void onConnect(WebSocket socket) {");
        pw.println("\t\tsuper.onConnect(socket);");
        if (eMethod.getParameters().size() == 1) {
            VariableElement ve = eMethod.getParameters().get(0);
            String remoteInterface = ve.asType().toString();
            String riclz = rp.getImplementationFor(remoteInterface);
            if (riclz == null) {
                pw.println("\t\tWebSocketWrapper wsw = WebSocketWrapper.getWebSocketWrapper(socket, this);");
            } else {
                checkRemoteMatching(clazz, riclz, rp);
                pw.println("\t\t\t"+riclz+" wsw = ("+riclz+") WebSocketWrapper.getWebSocketWrapper(socket, this);");
            }
        } else {
            throw new RuntimeException("onConnect method "+openMethodName+"must have one parameter");
        }
        
        
        
        //pw.println("\t\tWebSocketWrapper wsw = WSBeanServer.createWebSocketWrapper(socket);");
        pw.println("\t\tbean." + openMethodName + "(wsw);");
        pw.println("\t}"); 
    }
    
    static void writeOnMessageMethod(TypeElement clazz, 
                                    List<Element> methods, 
                                    RemoteProcessor rp, 
                                    PrintWriter pw) throws IOException {
        pw.println("\tpublic void onMessage(WebSocket socket, String messageString) {");
        //pw.println("\tSystem.out.println(\"START on message: \" +messageString);");
        //pw.println("\t\tString messageString = frame.getTextPayload();");
        for (Element nextMethod : methods) {
            MethodWriter.writeOnMessageCallbackBlockFor(clazz, nextMethod, rp, pw);
         }
         //pw.println("\tSystem.out.println(\"END on message: \" +messageString);");
         pw.println("\t}");
        
    }
    
    
    
    private static void writeOnMessageCallbackBlockFor(TypeElement clazz, 
                                                        Element method, 
                                                        RemoteProcessor rp, 
                                                        PrintWriter pw) throws IOException {
        ExecutableElement methodAsMethod = (ExecutableElement) method;
        WebSocketMessage wsmh = method.getAnnotation(WebSocketMessage.class);
        


        TypeMirror beanMethodReturnTM = methodAsMethod.getReturnType();
        boolean hasReturn = !(beanMethodReturnTM instanceof NoType);
        
        // suboptimal way of getting the classname of the firsr parameter...
        List parameters = methodAsMethod.getParameters();
        VariableElement firstParameter = (VariableElement) parameters.get(0);
        TypeMirror firstParameterTypeElement = firstParameter.asType();
        String firstParameterClassname = firstParameterTypeElement.toString();
        // end suboptimal way of...
        String endPointPath = getClassPath(clazz);
        pw.println("\t\tif (true) {");
        
                // declaration of web socket wrapper
        if (wsmh.requireremote()) {
            if (methodAsMethod.getParameters().size() < 2) {
                throw new RuntimeException("When you require a remote there would need to be 2 parameters....");
            }
            
            VariableElement ve = methodAsMethod.getParameters().get(1);
            String remoteInterface = ve.asType().toString();
            String riclz = rp.getImplementationFor(remoteInterface);

            if (riclz == null) {
                pw.println("\t\t\tWebSocketWrapper wsw = WebSocketWrapper.getWebSocketWrapper(socket, this);");
            } else {
                checkRemoteMatching(clazz, riclz, rp);
                pw.println("\t\t\t"+riclz+" wsw = ("+riclz+") WebSocketWrapper.getWebSocketWrapper(socket, this);");
            }

        } else {
            pw.println("\t\t\tWebSocketWrapper wsw = WebSocketWrapper.getWebSocketWrapper(socket, this);");
        }
        // end declaration of web socket wrapper
        //pw.println("System.out.println(\"in the on message wrapper impl\");");

        pw.println("\t\t\ttry {");
        String varNameForDecodedMessage = "decodedMessage";
        //System.out.println("-------" + firstParameterClassname);
        //pw.println("System.out.println(\"there is a decoder of type "+wsmh.decoder()+" \");");
        if (!wsmh.decoder().equals("")) {
            //pw.println("System.out.println(\"there is a decoder of type "+wsmh.decoder()+" \");");
            pw.println("\t\t\t\t" + wsmh.decoder() + " decoder = new " + wsmh.decoder() + "();");
            //pw.println("System.out.println(\"created "+wsmh.decoder()+" \");");
            pw.println("\t\t\t\tif (decoder.willDecode(messageString)) {");
            
            pw.println("\t\t\t\t\t" + firstParameterClassname + " " + varNameForDecodedMessage + " = decoder.decode(messageString);");
        } else {
            pw.println("\t\t\t\tif (true) {");
            String variableDeclarationForDecodedMessage = createVariableDeclarationForDecodedMessage(firstParameterClassname, varNameForDecodedMessage);
            pw.println("\t\t\t\t\t" + variableDeclarationForDecodedMessage);
        }
        
        String parametersToWrite = "";
        if (wsmh.requireremote()) {
            parametersToWrite = varNameForDecodedMessage + ", wsw";
        } else {
            parametersToWrite = varNameForDecodedMessage;
        }
        
        if (!hasReturn) { // this is the typical void case
            pw.println("\t\t\t\t\tbean." + method.getSimpleName() + "("+parametersToWrite + ");");
        } else {
            String returnClassname = methodAsMethod.getReturnType().toString();
            returnClassname = ProcessorUtils.convertPrimitiveToClassIfNec(returnClassname);
            pw.println("\t\t\t\t\t"+returnClassname+" returnMessage = bean." + method.getSimpleName() + "("+parametersToWrite + ");");
            pw.println("\t\t\t\t\tString returnMessageAsString = null;");
            if (!wsmh.encoder().equals("")) {
                pw.println("\t\t\t\t\t" + wsmh.encoder() + " encoder = new " + wsmh.encoder() + "();");
                pw.println("\t\t\t\t\treturnMessageAsString = encoder.encode(returnMessage);");
            } else if ( beanMethodReturnTM.toString().equals("byte") ||
                    beanMethodReturnTM.toString().equals("short")    ||
                    beanMethodReturnTM.toString().equals("int")      ||
                    beanMethodReturnTM.toString().equals("long")     ||
                    beanMethodReturnTM.toString().equals("float")    ||
                    beanMethodReturnTM.toString().equals("double")   ||
                    beanMethodReturnTM.toString().equals("boolean")  ||
                    beanMethodReturnTM.toString().equals("char") ) {
                
                pw.println("\t\t\t\t\treturnMessageAsString = returnMessage.toString();");
            } else if (beanMethodReturnTM.toString().equals("java.lang.String")) {
                //System.out.println("*******************************");
                pw.println("\t\t\t\t\treturnMessageAsString = (String) returnMessage;");
            } else {
                throw new RuntimeException("Can't handle this return type " + beanMethodReturnTM.toString());
            }

            
            pw.println("\t\t\t\t\twsw.sendMessage(returnMessageAsString);");  
        }
        pw.println("\t\t\t\t}");
        pw.println("\t\t\t} catch (Exception exception) {");
        pw.println("\t\t\t\tthis.handleException(exception);");
        pw.println("\t\t\t}");
        pw.println("\t\t}");
    }
    
    private static String createVariableDeclarationForDecodedMessage(String firstParameterClassname, String varNameForDecodedMessage) {
        String variableDeclarationForDecodedMessage = "";
        if (firstParameterClassname.equals("java.lang.String")) {
        // assume its a string
            variableDeclarationForDecodedMessage = "String "+varNameForDecodedMessage+" = messageString;";
            //pw.println("\t\t\tString "+varNameForDecodedMessage+" = messageString;");
        } else if (firstParameterClassname.equals("boolean")) {
            variableDeclarationForDecodedMessage = "boolean "+varNameForDecodedMessage+" = (new Boolean(messageString)).booleanValue();";
        } else if (firstParameterClassname.equals("int")) {
            variableDeclarationForDecodedMessage = "int "+varNameForDecodedMessage+" = (new Integer(messageString)).intValue();";
        } else if (firstParameterClassname.equals("short")) {
            variableDeclarationForDecodedMessage = "short "+varNameForDecodedMessage+" = (new Short(messageString)).shortValue();";
        } else if (firstParameterClassname.equals("long")) {
            variableDeclarationForDecodedMessage = "long "+varNameForDecodedMessage+" = (new Long(messageString)).longValue();";
        } else if (firstParameterClassname.equals("char")) {
            // FIXIT could use a more stringesnt test here rather than cherry picking the first character.
            variableDeclarationForDecodedMessage = "char "+varNameForDecodedMessage+" = messageString.charAt(0);";
        } else if (firstParameterClassname.equals("float")) {
            variableDeclarationForDecodedMessage = "float "+varNameForDecodedMessage+" = (new Float(messageString)).floatValue();";  
        } else if (firstParameterClassname.equals("double")) {
            variableDeclarationForDecodedMessage = "double "+varNameForDecodedMessage+" = (new Double(messageString)).doubleValue();";
         } else if (typeHasSingleStringConstructor(firstParameterClassname)) {
            variableDeclarationForDecodedMessage = firstParameterClassname+" "+varNameForDecodedMessage+" = new "+firstParameterClassname+"(messageString);";        
         } else if (!"".equals(typeHasFactoryMethod(firstParameterClassname)) ) {
            String methodName = typeHasFactoryMethod(firstParameterClassname);
            variableDeclarationForDecodedMessage = firstParameterClassname+" "+varNameForDecodedMessage+" = " + firstParameterClassname+"."+methodName+"(messageString);";   
        } else {
            System.out.println("AAAAAGH, ok it works , but don't do this !");
            variableDeclarationForDecodedMessage = "byte[] " +varNameForDecodedMessage+ " = messageString.getBytes();";

        }
        return variableDeclarationForDecodedMessage;
        
    }
    
    // return true for a class that can be loaded and that has a constructor with
    // a single string argument. e.g. new Thread(name);
    private static boolean typeHasSingleStringConstructor(String firstParameterClassname) {
        //System.out.println(" test " + firstParameterClassname);
        try {
            Class c = Class.forName(firstParameterClassname); // FIXIT think about the classloader for this...
            for (Constructor cons : c.getConstructors()) {
                if ( cons.getParameterTypes().length == 1 ) {
                    if (cons.getParameterTypes()[0].equals(java.lang.String.class)) {
                        return true;
                    }
                }
            }
            
        } catch (Throwable t) {
            System.out.println("Warning: cound not load " + firstParameterClassname);
        }
        
        return false;
    }
    
    private static String typeHasFactoryMethod(String firstParameterClassname) {
        //System.out.println(" test " + firstParameterClassname);
        try {
            Class c = Class.forName(firstParameterClassname); // FIXIT think about the classloader for this...
            for (Method method : c.getMethods()) {
                int modifiers = method.getModifiers();
                boolean isStatic = java.lang.reflect.Modifier.isStatic(modifiers);
                boolean isPublic = java.lang.reflect.Modifier.isPublic(modifiers);
                String name = method.getName();
                if ( !(isStatic && isPublic) ) {
                    return "";
                }
               
                if (name.equals("valueOf") ||
                    name.equals("fromString") ) {
                    if ( method.getParameterTypes().length == 1 ) {
                        if (method.getParameterTypes()[0].equals(java.lang.String.class)) {
                            return name;
                        }
                    }
                }
            }
            
        } catch (Throwable t) {
            System.out.println("Warning: cound not load " + firstParameterClassname);
        }
        
        return "";
    }
    
    static void writeOnCloseMethod(TypeElement clazz, Element method, RemoteProcessor rp, PrintWriter pw) {
        String closeMethodName = method.getSimpleName().toString();
        ExecutableElement eMethod = (ExecutableElement) method;
        pw.println("\tpublic void onClose(WebSocket socket) {");
        if (eMethod.getParameters().size() == 1) {
            VariableElement ve = eMethod.getParameters().get(0);
            String remoteInterface = ve.asType().toString();
            String riclz = rp.getImplementationFor(remoteInterface);
            if (riclz == null) {
                pw.println("\t\tWebSocketWrapper wsw = WebSocketWrapper.getWebSocketWrapper(socket, this);");
            } else {
                checkRemoteMatching(clazz, riclz, rp);
                pw.println("\t\t\t"+riclz+" wsw = ("+riclz+") WebSocketWrapper.getWebSocketWrapper(socket, this);");
            }
        } else {
            throw new RuntimeException("onOpen method "+closeMethodName+"must have one parameter");
        }
        pw.println("\t\tbean." + closeMethodName + "(wsw);");
        pw.println("\t}");
    }
    
    
}
