/*
 * To change this template, choose Tools | Templates and open the template in
 * the editor.
 */
package org.glassfish.websockets.platform.processors;



import org.glassfish.websockets.api.annotations.WebSocketRemote;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;

import javax.annotation.processing.*;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.*;
import javax.tools.Diagnostic;
import javax.tools.JavaFileObject;
import java.util.*;
/**
 *
 * @author dannycoward
 */

 
   
public class RemoteProcessor {
    private ProcessingEnvironment processingEnv;
    private Map<Element, Element> methodToClassMap;
    private Map<Element, List<Element>> classToMethodListMap;
    private Map<String, String> interfaceToImplementationFQClassnames = new HashMap<String, String>();
    
    public RemoteProcessor(ProcessingEnvironment processingEnv) {
        this.processingEnv = processingEnv;
    }
    
    public String getImplementationFor(String remoteInterfaceName) {
        //System.out.println("get implementation for: " + remoteInterfaceName);

        return interfaceToImplementationFQClassnames.get(remoteInterfaceName);
        
    }
    
    public static String getWrapperClassnameFor(String classname) {
        String className = classname.substring(classname.lastIndexOf(".")+1, classname.length());
        return className + "_RemoteImpl"; // FIXIT this should be based on long classnames
    }
    
    private boolean createRemoteImpls(Set annotations, RoundEnvironment roundEnv) {
        try {
            this.classToMethodListMap = new HashMap<Element, List<Element>>();
            for (Element nextClass : roundEnv.getRootElements()) {
                List<Element> aMethods = this.getAnnotatedMethods(nextClass);
                if (!aMethods.isEmpty()) {
                    this.checkImplementsWSRemote(nextClass);
                    classToMethodListMap.put(nextClass, aMethods);
                
                    //System.out.println("----" + classToMethodListMap);
                    String nextClazzQualName = ((TypeElement) nextClass).getQualifiedName().toString();
                    String wrapperPackage = ProcessorUtils.getWrapperPackageNameFor(nextClazzQualName);
                    String wrapperClassname = getWrapperClassnameFor(nextClazzQualName);
                    //System.out.println("->"+wrapperPackage);
                    //System.out.println("->"+wrapperClassname);
                    String wrapperFQName = wrapperPackage + "." + wrapperClassname;
                    interfaceToImplementationFQClassnames.put(nextClazzQualName, wrapperFQName);
                    JavaFileObject f = processingEnv.getFiler().
                                createSourceFile(wrapperFQName);

                    processingEnv.getMessager().printMessage(Diagnostic.Kind.NOTE,
                            "Creating " + f.toUri());
                    Writer w = f.openWriter();
                    try {
                        PrintWriter pw = new PrintWriter(w);
                        RemoteClassWriter.writePackageAndImports(pw, wrapperPackage);

                        pw.println();
                        RemoteClassWriter.writeClassDeclaration(pw, nextClazzQualName, wrapperClassname, aMethods);
                        pw.println();
                        pw.close();
                    } catch (Throwable t) {
                        t.printStackTrace();
                    }
                }
            }
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }  
        
        return true;
    }
    
    private void checkImplementsWSRemote(Element nextClass) {
        //System.out.println(nextClass);
        boolean ok = false;
        TypeElement te = (TypeElement) nextClass;
        List interfaces = te.getInterfaces();
        String wsRemoteName = org.glassfish.websockets.api.Peer.class.getName();
        for (TypeMirror nextInterface : te.getInterfaces()) {
            System.out.println(nextInterface.toString());
            if (wsRemoteName.equals(nextInterface.toString())) {
                ok = true;
            }
        }
        if (!ok) {
            throw new RuntimeException(nextClass + "does not implement WSRemote");
        }
    }
    
    private List<Element> getAnnotatedMethods(Element clazz) {
        List<Element> l = new ArrayList<Element>();
        if (clazz instanceof TypeElement) {
                for (Element annotatedThing : clazz.getEnclosedElements()) {
                    if (annotatedThing.getKind() == ElementKind.METHOD
                            && annotatedThing.getEnclosingElement() == clazz) {
                        WebSocketRemote remoteAn = annotatedThing.getAnnotation(WebSocketRemote.class);
                        if (remoteAn != null) {
                            l.add(annotatedThing);
                        }
                    }
                }
        }
        return l;
    }
    
    public boolean process(Set annotations, RoundEnvironment roundEnv) {
        return this.createRemoteImpls(annotations, roundEnv);
    }
    
    private Map<Element, Element> analyzeAnnotatedElements(Set annotations, 
                                                        RoundEnvironment roundEnv) {
        Map<Element, Element> methodToClassMap = new HashMap<Element, Element>();
        for (Element nextClass : roundEnv.getRootElements()) {
            if (nextClass instanceof TypeElement) {
                for (Element annotatedThing : nextClass.getEnclosedElements()) {
                    if (annotatedThing.getKind() == ElementKind.METHOD
                            && annotatedThing.getEnclosingElement() == nextClass) {
                        WebSocketRemote remoteAn = annotatedThing.getAnnotation(WebSocketRemote.class);
                        if (remoteAn != null) {
                            methodToClassMap.put(annotatedThing, nextClass);
                        }
                    }
                }
            }
        }
        return methodToClassMap;
    }
}
